program  Xdip_prog 
! Compute XPuck_dip (Fast) by using the 
!      xpuck_dip_init(aux)	
!      xpuck_dip_lop(data,residual)
! XPdip (Slow) is the same as Xdip program
!
! SYNTAX:  XPdip.x method=puck_f/s <*.H aux=*.H dip=*.H niter= eps1= eps2  nwind= v3d=T/F  verb=T/F maxsize=
!
! Author:  Mohammad Maysami
!          SEP, Stanford University
!          Jun 09

  use sep
  use sep3d_struct_mod
  use find_xz_dip_mod
  
  implicit none
  type(sep3d)       :: aux_s,data_s
  character(len=20) :: method
  ! real, pointer, dimension(:,:,:) :: xres3,dat3,aux3,dip3,coh3,res3
  real, pointer, dimension(:)       :: xres1,dat1,aux1,dip1
  real    :: eps1_dip,eps2_dip,icycle
  integer :: n(3),nwind(3),niter_dip,ndim,n123,n3,i3
  integer :: maxsize,ierr1,ierr2,stat
  logical :: verb,verb2,v3d
  ! logical :: dave

  call sep_init(SOURCE)
  call sep_begin_prog()

  call init_sep3d("in" ,data_s,"INPUT")
  call init_sep3d("aux",aux_s ,"INPUT")
  call from_param("verb",verb,.false.); 
  call from_param("v3d",v3d  ,.false.);  
  call from_param('max_size',maxsize,100); 
  maxsize=maxsize*250000

  ! Find Dip finding method
  method=" "
  call from_param("method",method,"sergey_f")
  write(0,*) "method=",method


  ! =================================================================
   ! Check size of mod and aux input
   n123=product(aux_s%n(1:sep3d_ndims(aux_s)))
   if (n123/=product(data_s%n(1:sep3d_ndims(data_s)))) &
        call seperr("Input and auxilary needs to be the same size")

  ! Get dimensions - 2D,3D
  ndim=sep3d_ndims(aux_s)
  if (ndim>3) then; 
     call erexit("can only calculate the dip on <= 3D data"); 
  end if
  n=1
  n(1:ndim)=aux_s%n(1:ndim)
  verb2=.false.
  if (.not. v3d) then; 
     n3=n(3); n(3)=1; 
     if (n3>1 .and. verb) then; 
        verb=.false.; 
        verb2=.true.; 
     end if; 
  else; 
     n3=1; 
  end if
  n123=product(n)
  icycle=.0001

  ! Set smoothing factors for SERGEY
  call from_param("niter",niter_dip,30) 
  call from_param("eps1",eps1_dip,0.00002)
  call from_param("eps2",eps2_dip,0.00002)

  ! Set smoothing factors for PUCK
  nwind(1)=min(n(1),5); 
  nwind(2)=min(n(2),5); 
  nwind(3)=min(n(3),5);       
  call from_param("nwind",nwind,nwind); 


  ! ==================  XLINE loop ======================================
  do i3=1,n3  
     ! Print progress
     if (verb2 .and. real(i3)/real(n3) > icycle ) then; 
        write(0,*) "working on ",nint(real(i3)/real(n3)*100.),"%"; icycle=icycle+.05; 
     end if !3d

     ! Check Max memory size
     if (n123>maxsize/4) then; 
        write(0,*) "max_size needs to be at least ",n123*16/1000000+1; 
        call seperr("maxsize is too small "); 
     end if

     ! Allocate in first try
     if (i3.eq.1)  &
          allocate(dat1(product(n)),aux1(product(n)),xres1(product(n)),dip1(product(n))) ! 1D Alloc.


     ! Read data and auxiliary data
     call sreed("aux",aux1,size(aux1)*4)
     call sreed("in",dat1,size(dat1)*4)

     ! Normalize : Done in xpuck_init     
     !      aux1=aux1/sum(abs(aux1))
     !      dat1=dat1/sum(abs(dat1))


     ! -------------------------------
     ! ------- FAST Puck Method ------
     if (method(1:6) .eq. "puck_f") then
        
        call xpuck_dip_init(aux_s,aux1,nwind)	
        stat=xpuck_dip_lop(.false.,.false.,dat1,xres1)
        call srite("out",xres1,size(xres1)*4)

        
     ! ----- SLOW Puck Method -----
     else if (method(1:6) .eq. "puck_s") then
        write(0,*) "Nothing to be done for method=",trim(method)
        stat=1

     ! ----- FAST Sergey Method -----
     else if (method(1:8) .eq. "sergey_f") then

        call xsergey_dip_init(aux_s,aux1,niter_dip,eps1_dip,eps2_dip)
        stat=xsergey_dip_lop(.false.,.false.,dat1,xres1)
        call srite("out",xres1,size(xres1)*4)

        
     ! ----- SLOW Sergeyk Method -----
     else if (method(1:8) .eq. "sergey_s") then

        write(0,*) "Nothing to be done for method=",trim(method)
        stat=1

     else
        ! Unknown methods
        stat=1
        write(0,*) "method=",trim(method); 
        call seperr( "Unknown method")
     end if

  end do 

  ! write out the dip
  if (stat==0) then
     call sep3d_write_description("dip",aux_s)
     call get_dip_aux(dip1)
     call srite("dip",dip1,size(dip1)*4)
  end if
     deallocate(dip1)
     


  deallocate(aux1,dat1,xres1)
  call sep_end_prog()

end program  
