module lpfilt_mod

  use sep
  use helix 
  use adj_mod

  implicit none
  type(filter), private  :: aa

contains
!   !-------------------------------------------------
!   subroutine lpf1_init(nd1,coeff)
!     optional :: coeff
!     real, dimension(:)              :: coeff ! Filter coefficients
!     real, dimension(:),allocatable  :: h ! Filter coefficients
!     integer                         :: i,nd1,nh   ! Ndata(1)

!     if (present(coeff)) then
!        nh = (size(coeff) )
!        allocate(h(nh))
!        h=coeff
!    else
!       nh=35
!       allocate(h(nh))
!       h=(/ 0.0042954 ,  0.00472237,  0.00597902,  0.00802558,  0.01079484,&
!          0.01419444,  0.01810997,  0.02240892,  0.02694525,  0.0315644 ,&
!          0.03610864,  0.04042254,  0.04435836,  0.04778116,  0.05057349,&
!          0.05263948,  0.05390817,  0.05433596,  0.05390817,  0.05263948,&
!          0.05057349,  0.04778116,  0.04435836,  0.04042254,  0.03610864,&
!          0.0315644 ,  0.02694525,  0.02240892,  0.01810997,  0.01419444,&
!          0.01079484,  0.00802558,  0.00597902,  0.00472237,  0.0042954 /)
!    end if


!     if (nh>nd1) then
!        call seperr('Length of Filter cannot be larger than n1 of data.')
!     end if

!    call allocatehelix(aa,nh)
   
!    do i=1,nh
!       aa%flt(i)=h(i)
!       aa%lag(i)=i-1
!    end do
!    deallocate(h)
!  end subroutine

  !-------------------------------------------------
  subroutine lpf_init(nd1,coeff)
    optional :: coeff
    real, dimension(:)              :: coeff ! Filter coefficients
    real, dimension(:),allocatable  :: h ! Filter coefficients
    integer                         :: i,nd1,nh   ! Ndata(1)

   if (present(coeff)) then
      nh = (size(coeff) )
      allocate(h(nh))
      h=coeff/sum(coeff)
   else
      nh=35
      allocate(h(nh))
      h=(/ 0.0042954 ,  0.00472237,  0.00597902,  0.00802558,  0.01079484,&
         0.01419444,  0.01810997,  0.02240892,  0.02694525,  0.0315644 ,&
         0.03610864,  0.04042254,  0.04435836,  0.04778116,  0.05057349,&
         0.05263948,  0.05390817,  0.05433596,  0.05390817,  0.05263948,&
         0.05057349,  0.04778116,  0.04435836,  0.04042254,  0.03610864,&
         0.0315644 ,  0.02694525,  0.02240892,  0.01810997,  0.01419444,&
         0.01079484,  0.00802558,  0.00597902,  0.00472237,  0.0042954 /)
      h=h/sum(h)
   end if

   call allocatehelix(aa,2*nh-1)
   
   aa%flt(1)=2*h(1)
   aa%lag(1)=0
   do i=2,nh
      aa%flt(i)=h(i)
      aa%lag(i)=i-1
      aa%flt(nh+i-1)=h(i)
      aa%lag(nh+i-1)=1+nd1*(i-1)
   end do
   deallocate(h)

 end subroutine

 !-------------------------------------------------
 function lpf_lop (adj,add,xx,yy)  result(stat)
   ! General Low-pass filter
   logical, intent(in) :: adj,add
   real,dimension(:)   :: xx,yy 
   integer             :: stat,ix,iy,ih,nn

   nn=size(xx)
   call adjnull (adj,add,xx,yy)
   if (adj) then
      do iy=1,nn
         do ih=1,size(aa%flt)
            ix= iy+aa%lag(ih)
            if (ix <= nn)  xx(ix) = xx(ix) + yy(iy)*aa%flt(ih)
         enddo
      enddo

   else
      do iy=1,nn
         do ih=1,size(aa%flt) 
            ix= iy+aa%lag(ih)
            if (ix <= nn) yy(iy) = yy(iy) + xx(ix)*aa%flt(ih);
         enddo
      enddo

   end if
   
   stat=0
 end function
 !-------------------------------------------------

 subroutine lpf_close() 

 call  deallocatehelix(aa)
 end subroutine

end module
