module tomo_reg_prec_mod
use sep
use nhelicon2
use npolydiv2
use nhelicon
use tomo_eps_mod
use npolydiv
use helicon
use chain_mod
use sep3d_struct_mod
use io_nfilt
use xgrad_mod
use find_xz_dip_mod 
use lpfilt_mod
use identity_mod

implicit none
logical, private, save :: prec,dual,new
integer, private, save :: nmod,xflag,smooth(3)
real, private,save,pointer :: space(:),space2(:)
real, private, save :: coeff(10)

type(nfilter) :: filt1_prec,filt2_prec
type(nfilter) :: filt1_reg,filt2_reg,filta,filtb
type(filter)  :: filt1


contains

! --------------------------------------
logical function load_reg_prec(slow_s,reg0) 
  type(sep3d)       :: slow_s,reg_s,aux_s		      
  real   , pointer, dimension(:)     :: reg0,junk,xaux0,aux1,temp1
  real   , pointer, dimension(:,:,:) :: aux3
  integer, external :: auxin,sreed
  integer :: ierr,i,ns,niter_dip,nd1,nd2
  real    :: sm,eps1_dip,eps2_dip
  
  load_reg_prec=.false.
  
  call from_param("new",new,.true.)
  call from_param("prec",prec,.true.)
  call from_param("xflag",xflag,0)

  ! Set smoothing factors for Dip
  call from_param("niter_dip",niter_dip,30) 
  call from_param("eps1_dip",eps1_dip,0.00002)
  call from_param("eps2_dip",eps2_dip,0.00002)
  !   call from_param("smooth",smooth,(/5,5,1/))

  write (0,*), "---------- TOMO REG/PREC FLAGS ----------"
  write (0,*), " - TOMO NEW   FLAG:",new
  write (0,*), " - TOMO PREC  FLAG:",prec
  write (0,*), " - TOMO XProp FLAG:",xflag

  write (0,*), " - NITER DIP:",niter_dip
  write (0,*), " - EPS1  DIP:",eps1_dip
  write (0,*), " - EPS2  DIP:",eps2_dip

  nmod=product(slow_s%n(1:sep3d_ndims(slow_s)))
  nd1=slow_s%n(1)
  nd2=slow_s%n(2)
  allocate(reg0(nmod))


  if (xflag /= 0) then

     ! Set and Read AUX Soft-data
     ! ==========================
     allocate(aux1(nmod))

     if(-1==auxin("aux")) then
        write (0,*) " aux tag is not provided. Initialization failed !"
        return
     else
        call init_sep3d("aux",aux_s,"INPUT")
        if(product(aux_s%n) /= size(aux1)) then
           write(0,*) "Size of aux and the slowness space must be equal !"
           return
        end if
        
        
        if(4*size(aux1)/=sreed("aux",aux1,size(aux1)*4)) then
           write(0,*) "Trouble reading in aux !"
           return
        end if
        
     end if

     if (xflag==1) then
        call xgrad_init(aux_s,aux1)
     elseif (xflag >=2) then
        ! Init both XPuck and LPF  A=XP.LPF.(s+ds)
        call xsergey_dip_init(aux_s,aux1,niter_dip,eps1_dip,eps2_dip)
        ! Freq_cut = 2/n * Freq.
        coeff=(/0.01481481,  0.03108374,  0.07367633,  0.12632367,  0.16891626,& 
             0.18518519,  0.16891626,  0.12632367,  0.07367633,  0.03108374/)
        call lpf_init(nd1,coeff)
     endif


     !       XAUX0 Part
     ! ==========================
     allocate(xaux0(nmod))
     ! allocate(temp1(nmod))

     if(-1==auxin("xaux0")) then
        xaux0=0.
        ! write (0,*) "Tomo_REG: Xaux0 computed by code."
        ! ===> SLOW needs to be added if these are uncommented.
        ! if (xflag==1) then
        !   call xgrad_init(aux_s,aux1)
        !   ierr=xgrad_lop (.false.,.false.,slow,xaux0)
        ! elseif (xflag==2) then
        !   ! Init both XPuck and LPF  A=XP.LPF.(s+ds)
        !   call xsergey_dip_init(aux_s,aux1,niter_dip,eps1_dip,eps2_dip)
        !   call lpf_init(nd1)       
        !   call chain(xsergey_dip_lop,lpf_lop,adj,add,model,data,space)
        !   !ierr=xpsergey_dip_lop(.false.,.false.,slow,xaux0)
        !   call lpf_init(nd1)
        !   call chain(xpsergey_dip_lop,lpf_op,.false.,.false.,slow,xaux0,temp1)
        ! endif
        ! xaux0=-xaux0
     else
        call init_sep3d("xaux0",aux_s,"INPUT")
        if(product(aux_s%n) /= size(xaux0)) then
           write(0,*) "Size of xaux0 and the slowness space must be equal !"
           return
        end if
        
        if(4*size(xaux0)/=sreed("xaux0",xaux0,size(xaux0)*4)) then
           write(0,*) "trouble reading in xaux0"
           return
        end if
        ! call sep3d_write_description("check_xaux0.H",aux_s)
        ! call srite("check_xaux0.H",xaux0,size(xaux0)*4)
        xaux0=-xaux0 ! XG(s0+ds)=0 ==> XG.ds= -XG.s0
     end if
	
     call sep3d_write_description("check_xaux0.H",aux_s)
     call srite("check_xaux0.H",xaux0,size(xaux0)*4)


  end if  !! XFLAG END
  !#! else  ! XFLAG=FALSE 
  ! --------------------------------------------------------------





     !       REG0 Part
     ! ==========================
     if(-1==auxin("reg0")) then
        reg0=0.

     else
        call init_sep3d("reg0",reg_s,"INPUT")
        if(product(reg_s%n) /= size(reg0)) then
           write(0,*) "Size of reg0 and the slowness space must be equal !"
           return
        end if
    		
        if(4*size(reg0)/=sreed("reg0",reg0,size(reg0)*4)) then
           write(0,*) "trouble reading in reg0"
           return
        end if
        ! call sep3d_write_description("check_reg0.H",reg_s)
        ! call srite("check_reg0.H",reg0,size(reg0)*4)
        
        reg0=-reg0


        ns=0; sm=0
        do i=1,size(reg0)
           if(abs(reg0(i)) <.3.and. i < 200) then
              sm=sm+reg0(i)
              ns=ns+1
           end if
        end do
        
        sm=sm/ns
        ! sm=.06
        
        do i=1,size(reg0)
           if(abs(reg0(i)) <.3 .and. i < 200) then
              reg0(i)=reg0(i)-sm
              ! write(0,*) "SUBTA",reg0(i),sm
           else
              reg0(i)=0
           end if
        end do
		
        ! write(0,*) "CHECK SM",sm,maxval(reg0),minval(reg0)
     end if
	
     call sep3d_write_description("check_reg0.H",reg_s)
     call srite("check_reg0.H",reg0,size(reg0)*4)


     !  ==== REGULARIZATION Part 1 ====
     if(0==auxin("filt_reg")) then
        write(0,*) "filt not supplied"
        return
     end if
     
     if(.not. read_nfilt("filt_reg",filt1_reg)) then
        write(0,*) "trouble reading nfilter"
        return
     end if
	
     dual=.false.
     if(-1/=auxin("filt2_reg")) then
        dual=.true.
        if(.not. read_nfilt("filt2_reg",filt2_reg)) then
           write(0,*) "trouble reading nfilter"
           return
        end if
        ! allocate(space(nmod))
        
     end if

     ! --- NEW FLAG ---
     if (new) then
        if(.not. read_nfilt("filt1",filtb)) then
           write(0,*) "trouble reading nfilter"
           return
        end if
        
        if(.not. read_nfilt("filt2",filta)) then
           write(0,*) "trouble reading nfilter"
           return
        end if
     end if

     allocate(space(nmod),space2(nmod))


  !  ==== REGULARIZATION Part 2 ====
  call allocatehelix(filt1,3)
  ierr=slow_s%n(1)
  filt1%lag=(/ierr-1,ierr,ierr/)
  filt1%flt=(/-.33,-.33,-.33/)
  call helicon_init(filt1)
  
  ! --- FALSE NEW FLAG ---
  if(.not. new) then
     call nhelicon_init(filt1_reg)
     if(dual) call nhelicon2_init(filt2_reg)
     
     if(prec) then
        call npolydiv_init(nmod,filt1_prec)
        if(dual) call npolydiv2_init(nmod,filt2_prec)
     end if
  else
     call npolydiv_init(nmod,filta)
     call nhelicon_init(filtb)
  end if
  
!!!!   end if  ! XFLAG END


  !     PRE-Cond. Part
  ! ==========================
  if(prec) then
     if(.not. read_nfilt("filt_prec",filt1_prec)) then
        write(0,*) "trouble reading nfilter"
        return
     end if
     
     dual=.false.
     if(-1/=auxin("filt2_prec")) then
        dual=.true.
        if(.not. read_nfilt("filt2_prec",filt2_prec)) then
           write(0,*) "trouble reading nfilter"
           return
        end if
        !   allocate(space(nmod))
     end if
  end if



  if (xflag /= 0) reg0=xaux0
  load_reg_prec=.true.
end function 


! --------------------------------------
integer function reg_op(adj,add,model,data)
  logical, intent(in) :: adj,add
  real :: model(:),data(:)
  integer :: ierr
  
  reg_op=0;
  call from_param("xflag",xflag,0)

  if (xflag==1) then
     ! write(0,*) "REG_OP: XGRAD is used"
     ierr=xgrad_lop (adj,add,model,data)

  elseif (xflag==2) then
     ! write(0,*) "REG_OP: XDIP is used"
     ierr=xsergey_dip_lop(adj,add,model,data)     


  elseif (xflag==3) then
     ! write(0,*) "REG_OP: XDIP+LPF is used"
     call chain(xsergey_dip_lop,lpf_lop,adj,add,model,data,space)
     !!!! call chain(xsergey_dip_lop,id_lop,adj,add,model,data,space)


  else
     ! write(0,*) "REG_OP: REG is used"
     if(new) then
        !write(0,*) "in reg",adj
        ! call chain(nhelicon_lop,prec_op,adj,add,model,data,space2)
        ierr=nhelicon_lop(adj,add,model,data)
     else
        if(dual) then
           call chain(nhelicon_lop,nhelicon2_lop,adj,add,model,data,space)
        else
           ierr=nhelicon_lop(adj,add,model,data)
           ! ierr=helicon_lop(adj,add,model,data)
        end if
     end if
  end if
end function reg_op



! --------------------------------------
integer function prec_op(adj,add,model,data)
  logical, intent(in) :: adj,add
  real :: model(:),data(:)
  prec_op=0
  !   write (0,*) "PREC Operator Called"
  ! ---NEW FLAG---
  if(new) then
     !call chain(npolydiv_lop,nhelicon_lop,adj,add,model,data,space)
     ! call adjnull(adj,add,model,data)
     if(1==1) then
        prec_op=npolydiv_lop(adj,add,model,data)
     endif
     
     !write(0,*) "BB",adj
     if(1==2) then
        if(.not. adj) then
           data=model+data
        else
           model=data+model
        end if
     end if
  else
     if (prec) then
        if(dual) then
           call chain(npolydiv_lop,npolydiv2_lop,adj,add,model,data,space)
        else
           prec_op=npolydiv_lop(adj,add,model,data)
        end if
     else 
        if(.not. adj) then
           data=model+data
        else
           model=data+model
        end if
     end if
  end if
end function prec_op

  

! --------------------------------------
integer function reg_range()
  reg_range=nmod
end function

integer function reg_domain()
  reg_domain=nmod
end function

integer function prec_range()
  prec_range=nmod
end function

integer function prec_domain()
  prec_domain=nmod
end function


! --------------------------------------

end module
