! Wave Kill Module
! Author:  Mohammad Maysami
!          SEP, Stanford University
!          Jun 09

module wavekill_mod 

  use adj_mod


implicit none
real, dimension(:,:),private, pointer  :: s11,s12,s21,s22
integer, private :: nt,nx

contains 
 
! ----------------------------------------------
subroutine wavekill_init(aa,bb,n1,n2)
  ! Initialize wavekill function
  ! Implements  v=(a*d/dx + b*d/dt) u
  ! P* d/dt= [-1 -1; 1 1]
  ! 1* d/dx= [-1  1;-1 1]
  real    :: aa,bb(:,:)
  integer :: n1,n2 

  allocate(s11(n1,n2),s12(n1,n2),s21(n1,n2),s22(n1,n2))
  nt=n1; nx=n2

  ! Form wavekill coefficients
  s11 = -aa-bb; s12 = aa-bb 
  s21 = -aa+bb; s22 = aa+bb
end subroutine


! =============================================
function wavekill_lop (adj,add,mm,dd) result(stat)
  logical, intent(in)         :: adj,add
  real,dimension(:)           :: mm,dd 
  integer                     :: stat

  call adjnull (adj,add,mm,dd)  
  call wavekill_lop2(adj,add,mm,dd)
  stat=0

end function


! =============================================
subroutine wavekill_lop2(adj,add,uu,vv)
  ! Implements  v=(a*d/dx + b*d/dt) u
  logical, intent(in)    :: adj,add
  real, dimension(nt,nx) :: uu ,vv
  integer                :: it,ix

  if (adj) then
     do ix=1,nx-1
        do it=1,nt-1
           uu(it  ,ix  ) = uu(it  ,ix  )+ s11(it,ix)*vv(it,ix) 
           uu(it  ,ix+1) = uu(it  ,ix+1)+ s12(it,ix)*vv(it,ix)
           uu(it+1,ix  ) = uu(it+1,ix  )+ s21(it,ix)*vv(it,ix)
           uu(it+1,ix+1) = uu(it+1,ix+1)+ s22(it,ix)*vv(it,ix)
        end do
     end do
     !!!!! BOUNDARIES EXCLUDED

!      vv(nt,ix) = vv(nt,ix)+                  ! vv(nt-1,ix) 
!                 uu(nt-1  ,ix  )*s11(nt-1,ix)+& 
!                 uu(nt-1  ,ix+1)*s12(nt-1,ix)+& 
!                 uu(nt-1+1,ix  )*s21(nt-1,ix)+& 
!                 uu(nt-1+1,ix+1)*s22(nt-1,ix) 
!      vv(it,nx) = vv(it,nx)+ vv(it,nx-1) 
!                 uu(it  ,nx-1  )*s11(it,nx-1)+& 
!                 uu(it  ,nx-1+1)*s12(it,nx-1)+& 
!                 uu(it+1,nx-1  )*s21(it,nx-1)+& 
!                 uu(it+1,nx-1+1)*s22(it,nx-1) 

!      do ix=1,nx-1
!         uu(nt-1  ,ix  ) = uu(nt-1  ,ix  ) + s11(nt-1,ix) * vv(nt,ix) 
!         uu(nt-1  ,ix+1) = uu(nt-1  ,ix+1) + s12(nt-1,ix) * vv(nt,ix) 
!         uu(nt-1+1,ix  ) = uu(nt-1+1,ix  ) + s21(nt-1,ix) * vv(nt,ix) 
!         uu(nt-1+1,ix+1) = uu(nt-1+1,ix+1) + s22(nt-1,ix) * vv(nt,ix) 
!      end do
!      do it=1,nt-2   
!         uu(it  ,nx-1  ) = uu(it  ,nx-1  ) + s11(it,nx-1)*vv(it,nx)
!         uu(it  ,nx-1+1) = uu(it  ,nx-1+1) + s12(it,nx-1)*vv(it,nx)
!         uu(it+1,nx-1  ) = uu(it+1,nx-1  ) + s21(it,nx-1)*vv(it,nx)
!         uu(it+1,nx-1+1) = uu(it+1,nx-1+1) + s22(it,nx-1)*vv(it,nx)
!      end do

  else
     do ix=1,nx-1
        do it=1,nt-1 
           vv(it,ix)=vv(it,ix)+&
                uu(it  ,ix  )*s11(it,ix)+& 
                uu(it  ,ix+1)*s12(it,ix)+& 
                uu(it+1,ix  )*s21(it,ix)+& 
                uu(it+1,ix+1)*s22(it,ix) 
        end do
     end do
     vv(nt,:) = vv(nt,:)+ 0. !vv(nt-1,:) 
     vv(:,nx) = vv(:,nx)+ 0. !vv(:,nx-1) 

  end if

end subroutine


! =============================================
subroutine wavekill_close ()
  
  deallocate(s11,s12,s21,s22)

end subroutine



! ! =============================================
! subroutine wavekill_org(aa,bb,nt,nx,uu,vv)
!   ! Implements  v=(a*d/dx + b*d/dt) u
!   ! d/dt= [-1 -1; 1 1]
!   ! d/dx= [-1  1;-1 1]
!   real    :: aa,bb(:,:),uu(:,:),vv(:,:) 
!   integer :: it,ix,nt,nx 
!   real    :: s11(nt,nx),s12(nt,nx),s21(nt,nx),s22(nt,nx) 
  
!   s11 = -aa-bb; s12 = aa-bb 
!   s21 = -aa+bb; s22 = aa+bb
!   vv=0. 

!   do ix=1,nx-1
!      do it=1,nt-1 
!         vv(it,ix)=uu(it ,ix )*s11(it,ix)+& 
!              uu(it ,ix+1)*s12(it,ix)+& 
!              uu(it+1,ix )*s21(it,ix)+& 
!              uu(it+1,ix+1)*s22(it,ix) 
!      end do
!   end do
!   vv(nt,:)=vv(nt-1,:) 
!   vv(:,nx)=vv(:,nx-1) 
! end subroutine

end module 
